"""ENTITIES ADVANCED
Definition of advanced classes based on the base entities
"""

# # Native # #
from datetime import datetime, timedelta, tzinfo
from typing import Optional, List

# # Package # #
from .entities import Stop, Bus, Stops, BusTime
from .helpers import *

__all__ = ("AdvancedStop", "AdvancedBus")


class AdvancedStop(Stop):
    """Stop to where Buses will arrive.
    This AdvancedStop has more attributes than the base Stop:
    - created: datetime when this Stop was first registered
    - updated: datetime when this Stop was last updated
    - original_name: original Stop name from an external data source
    - tags: list of extra tags for this Stop
    - extra_names: list of extra names for this Stop
    - source: name of the data source this data was fetched from
    """
    created: Optional[datetime]
    updated: Optional[datetime]
    original_name: Optional[str]
    tags: List[str] = []
    extra_names: List[str] = []
    distance: Optional[float]
    source: Optional[str]


class AdvancedBus(Bus):
    """This AdvancedBus has more attributes than the base Bus:
    - arrival: absolute or relative time until arrival to the Stop
    - departure: absolute or relative time until departure from the Stop
    - stops: list of stops this bus will stop at (used for Static bus definitions)
    """
    arrival: BusTime
    departure: BusTime
    stops: Stops

    def relative_arrival(self, timezone: Optional[tzinfo]) -> timedelta:
        """Get the time difference between current time and the 'arrival' of this Bus (must be a relative datetime).
        A custom timezone can be used for the current time, otherwise default 'datetime' parameter is system timezone.
        """
        return calculate_relative_time(self.arrival, timezone)

    def relative_departure(self, timezone: Optional[tzinfo]) -> timedelta:
        """Get the time difference between current time and the 'departure' of this Bus (must be a relative datetime).
        A custom timezone can be used for the current time, otherwise default 'datetime' parameter is system timezone.
        """
        return calculate_relative_time(self.departure, timezone)

    def time_in_stop(self) -> timedelta:
        """Get the difference of time between the departure and the arrival time"""
        return self.departure - self.arrival



"""EXCEPTIONS
Custom exceptions for issues related with the Stops and Buses
"""


class PyBusesException(Exception):
    """Base exception for all the custom PyBusesEntieies exceptions
    """
    pass


class StopException(PyBusesException):
    """Base exception for all the Stop-related exceptions
    """
    pass


class StopNotFound(StopException):
    """Exception raised when a Stop is Not Found on a certain data source,
    but it might exist on other data sources
    """
    pass


class StopNotExist(StopException):
    """Exception raised when a trustful data source confirms that a Stop not exist physically
    """
    pass


class BusException(PyBusesException):
    """Base exception for all the Bus-related exceptions
    """
    pass



"""SORTING
Functions to sort lists of objects by one of their attributes (they can be provided as 'key' of sorted/sort methods)
"""

__all__ = ("BusSort", "StopSort")


class BusSort:
    """Keys to sort lists of Buses
    """

    @staticmethod
    def line(bus):
        """Key to sort a list of buses by their Line
        """
        return bus.line

    @staticmethod
    def route(bus):
        """Key to sort a list of buses by their Route
        """
        return bus.route

    @staticmethod
    def line_route(bus):
        """Key to sort a list of buses by their Line, then by their Route
        """
        return bus.line, bus.route

    @staticmethod
    def time(bus):
        """Key to sort a list of buses by their Time
        """
        return bus.time

    @staticmethod
    def time_line(bus):
        """Key to sort a list of buses by their Time, then by their Line
        """
        return bus.time, bus.line

    @staticmethod
    def time_route(bus):
        """Key to sort a list of buses by their Time, then by their Route
        """
        return bus.time, bus.route

    @staticmethod
    def time_line_route(bus):
        """Key to sort a list of buses by their Time, then by their Line, then by their Route
        """
        return bus.time, bus.line, bus.route

    @staticmethod
    def arrival(bus):
        """Key to sort a list of buses by their Arrival Time
        """
        return bus.arrival

    @staticmethod
    def departure(bus):
        """Key to sort a list of buses by their Departure Time
        """
        return bus.departure


class StopSort:
    """Keys to sort lists of Stops
    """

    @staticmethod
    def stopid(stop):
        """Key to sort a list of stops by their Stop ID
        """
        return stop.stopid

    @staticmethod
    def distance(stop):
        """Key to sort a list of stops by their Distance
        """
        return stop.distance



"""HELPERS
Helper functions for class methods
"""

# # Native # #
import hashlib
import copy
from typing import Optional, Dict
from datetime import datetime, timedelta, tzinfo

__all__ = ("calculate_relative_time", "generate_busid", "clear_dict_values")


def calculate_relative_time(timestamp: datetime, timezone: Optional[tzinfo] = None) -> timedelta:
    """Get the time difference between current time and the given time.
    A custom timezone can be used for the current time, otherwise default 'datetime' parameter is system timezone.
    """
    return datetime.now(tz=timezone) - timestamp


def generate_busid(line: str, route: str) -> str:
    """Generate a Bus ID given its line and route.
    Bus ID is generated as a MD5 checksum from the line and route, and returned as string.
    If line or route not exist, they will be ignored. Is both not exist, all Buses will have the same Bus ID.
    """
    h = hashlib.md5()
    if line:
        h.update(line.encode())
    if route:
        h.update(route.encode())
    return h.hexdigest()


def clear_dict_values(
        d: Dict,
        remove_none: bool,
        remove_empty_strings: bool,
        remove_empty_lists: bool,
        remove_empty_dicts: bool
) -> Dict:
    """Given a Dict, clear the Values that are null (None) or empty strings, depending on the function params.
    A copy of the given dict is returned.
    """
    cloned_dict = copy.deepcopy(d)
    for key, value in d.items():
        conditions = (
            remove_none and value is None,
            remove_empty_strings and value == "",
            remove_empty_lists and isinstance(value, list) and not value,
            remove_empty_dicts and isinstance(value, dict) and not value
        )
        if any(conditions):
            cloned_dict.pop(key)
        elif isinstance(value, dict):
            cloned_dict[key] = clear_dict_values(
                value, remove_none, remove_empty_strings, remove_empty_lists, remove_empty_dicts
            )

    return cloned_dict



#!/usr/bin/env python

from setuptools import setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="pybuses-entities",
    version="0.0.2",
    description="Data classes to work with Stops and Buses/vehicles on a public transportation network",
    author="David Lorenzo",
    url="https://github.com/David-Lor/PyBuses",
    packages=("pybusent",),
    install_requires=("pydantic",),
    license="Apache 2.0",
    classifiers=(
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ),
    long_description_content_type="text/markdown",
    long_description=long_description,
)



"""ENTITIES
Definition of classes & typing definitions
"""

# # Native # #
from typing import List, Optional, Union, Dict
from datetime import datetime, timedelta, tzinfo

# # Installed # #
import pydantic

# # Package # #
from .helpers import *

__all__ = ("Stop", "Bus", "BusTime", "Buses", "Stops", "BusesResult")


BusTime = Optional[Union[int, float, datetime, timedelta]]
"""Typing for the Time remaining or scheduled arrival/departure of a Bus. 
Can be an absolute datetime object, or the relative time as int or float
"""

Timestamp = Optional[Union[str, int, float, datetime]]
"""Typing for timestamps used on Stop created/updated values. 
Should be an absolute timestamp in any format prefered: datetime, str representation, or int/float epoch timestamp"""


class BaseEntity(pydantic.BaseModel):
    """Base Entity for all the PyBuses Entities
    """

    def get_dict(
            self,
            remove_none: bool = True,
            remove_empty_strings: bool = True,
            remove_empty_lists: bool = True,
            remove_empty_dicts: bool = True
    ) -> Dict:
        """Return the dict representation of the current object, with value cleaning.
        :param remove_none: if True, remove items with null/None value (default: True)
        :param remove_empty_strings: if True, remove empty-string items (default: True)
        :param remove_empty_lists: if True, remove empty lists (default: True)
        :param remove_empty_dicts: if True, remove empty dicts (default: True)
        """
        return clear_dict_values(dict(self), remove_none, remove_empty_strings, remove_empty_lists, remove_empty_dicts)

    @staticmethod
    def get_sort_key(attribute_name: str):
        """Get a sorting key based on the given attribute name
        """
        return lambda inst: inst.__getattr__(attribute_name)


class Stop(BaseEntity):
    """Stop to where Buses will arrive:
    - stopid: ID or code number of this Stop (required, int only)
    - name: canonical name of this stop (required)
    - lat, lon: location of this stop (latitude & longitude)
    """
    stopid: int
    name: str
    lat: Optional[float]
    lon: Optional[float]

    def has_location(self) -> bool:
        """Return True if this Stop has a location defined
        """
        return (self.lat is not None) and (self.lon is not None)


Stops = List[Stop]
"""Typing for a List of Stops. Usually a list of stops for a certain static bus line-route definition"""


class Bus(BaseEntity):
    """Vehicle that arrive to a Stop at or in a certain time:
    - busid: autogenerated if not specified, based on line and route
    - line: Bus line (required)
    - route: Bus route (required)
    - time: generic absolute or relative time, usually until arrival to the Stop.
            To differentiate between arrival and departure times, use the AdvancedBus
            with the arrival and departure attributes.
    """
    busid: Optional[str]
    line: str
    route: str
    time: BusTime

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.busid is None:
            self.busid = generate_busid(self.line, self.route)

    def relative_time(self, timezone: Optional[tzinfo]) -> timedelta:
        """Get the time difference between current time and the 'time' of this Bus (which must be a relative datetime).
        A custom timezone can be used for the current time, otherwise default 'datetime' parameter is system timezone
        """
        return calculate_relative_time(self.time, timezone)


Buses = List[Bus]
"""Typing for List (Array) of Buses. Usually a list of buses that will come to a Stop"""


class BusesResult(BaseEntity):
    """A contextualized Result of Buses Query, with useful fields to get more context of the result:
    - buses: List of Buses (default: empty list)
    - more_buses_available: True if more buses were available on the data source, but not all of them were returned
                            on the current query
    - stop: optional Stop that can be appended to a Buses Result Query, when the data source returns Stop info
            alongside the Buses list
    - source: name of the data source this data was fetched from
    """
    buses: Buses = []
    stop: Optional[Stop]
    more_buses_available: Optional[bool]
    source: Optional[str]
